@@ -204,16 +204,15 @@ delete(Section, Key, Persist, Reason) when is_list(Section), is_list(Key) ->
 
 
 features() ->
-    lists:usort([list_to_atom(Key) || {Key, "true"} <- ?MODULE:get(?FEATURES)]).
-
+    application:get_env(config, enabled_features, []).
 
 enable_feature(Feature) when is_atom(Feature) ->
-    ?MODULE:set(?FEATURES, atom_to_list(Feature), "true", false).
-
+    application:set_env(config, enabled_features,
+        lists:usort([Feature | features()]), [{persistent, true}]).
 
 disable_feature(Feature) when is_atom(Feature) ->
-    ?MODULE:delete(?FEATURES, atom_to_list(Feature), false).
-
+    application:set_env(config, enabled_features,
+        features() -- [Feature], [{persistent, true}]).
 
 listen_for_changes(CallbackModule, InitialState) ->
     config_listener_mon:subscribe(CallbackModule, InitialState).
@@ -301,17 +300,29 @@ handle_call(reload, _From, Config) ->
     end, dict:new(), Config#config.ini_files),
     % Update ets with anything we just read
     % from disk
-    dict:fold(fun(K, V, _) ->
-        ets:insert(?MODULE, {K, V})
+    dict:fold(fun({Sec, Key} = K, V, _) ->
+        VExisting = get(Sec, Key, V),
+        ets:insert(?MODULE, {K, V}),
+        case V =:= VExisting of
+            true ->
+                ok;
+            false ->
+                couch_log:notice("Reload detected config change ~s.~s = ~p", [Sec, Key, V]),
+                Event = {config_change, Sec, Key, V, true},
+                gen_event:sync_notify(config_event, Event)
+        end
     end, nil, DiskKVs),
     % And remove anything in ets that wasn't
     % on disk.
-    ets:foldl(fun({K, _}, _) ->
+    ets:foldl(fun({{Sec, Key} = K, _}, _) ->
         case dict:is_key(K, DiskKVs) of
             true ->
                 ok;
             false ->
-                ets:delete(?MODULE, K)
+                couch_log:notice("Reload deleting in-memory config ~s.~s", [Sec, Key]),
+                ets:delete(?MODULE, K),
+                Event = {config_change, Sec, Key, deleted, true},
+                gen_event:sync_notify(config_event, Event)
         end
     end, nil, ?MODULE),
     {reply, ok, Config}.
