@@ -25,6 +25,7 @@
 
 
 -define(TIMEOUT, 4000).
+-define(RESTART_TIMEOUT_IN_MILLISEC, 3000).
 
 -define(CONFIG_FIXTURESDIR,
         filename:join([?BUILDDIR(), "src", "config", "test", "fixtures"])).
@@ -183,8 +184,9 @@ config_features_test_() ->
             fun setup/0,
             fun teardown/1,
             [
-                fun should_enable_features/0,
-                fun should_disable_features/0
+                {"enable", fun should_enable_features/0},
+                {"disable", fun should_disable_features/0},
+                {"restart config", fun should_keep_features_on_config_restart/0}
             ]
         }
     }.
@@ -291,7 +293,23 @@ config_notifier_behaviour_test_() ->
                 {["section_foo"], fun should_not_notify/2},
                 {[{"section_foo", "key_bar"}], fun should_not_notify/2},
                 {all, fun should_unsubscribe_when_subscriber_gone/2},
-                {all, fun should_not_add_duplicate/2}
+                {all, fun should_not_add_duplicate/2},
+                {all, fun should_notify_on_config_reload/2},
+                {all, fun should_notify_on_config_reload_flush/2}
+            ]
+        }
+    }.
+
+
+config_key_has_regex_test_() ->
+    {
+        "Test key with regex can be compiled and written to file",
+        {
+            foreach,
+            fun setup/0,
+            fun teardown/1,
+            [
+                fun should_handle_regex_patterns_in_key/0
             ]
         }
     }.
@@ -318,6 +336,11 @@ should_write_config_to_file() ->
     ?assertEqual(ok, config:set("admins", "foo", "500", true)).
 
 
+should_handle_regex_patterns_in_key() ->
+    ?assertEqual(ok, config:set("sect1", "pat||*", "true", true)),
+    ?assertEqual([{"pat||*", "true"}], config:get("sect1")).
+
+
 should_delete_config_from_file() ->
     ?assertEqual(ok, config:delete("admins", "foo", true)).
 
@@ -651,6 +674,7 @@ should_not_add_duplicate(_, _) ->
 
 
 should_enable_features() ->
+    [config:disable_feature(F) || F <- config:features()],
     ?assertEqual([], config:features()),
 
     ?assertEqual(ok, config:enable_feature(snek)),
@@ -664,6 +688,7 @@ should_enable_features() ->
 
 
 should_disable_features() ->
+    [config:disable_feature(F) || F <- config:features()],
     ?assertEqual([], config:features()),
 
     config:enable_feature(snek),
@@ -675,6 +700,35 @@ should_disable_features() ->
     ?assertEqual(ok, config:disable_feature(snek)),
     ?assertEqual([], config:features()).
 
+should_keep_features_on_config_restart() ->
+    [config:disable_feature(F) || F <- config:features()],
+    ?assertEqual([], config:features()),
+
+    config:enable_feature(snek),
+    ?assertEqual([snek], config:features()),
+    with_process_restart(config),
+    ?assertEqual([snek], config:features()).
+
+should_notify_on_config_reload(Subscription, {_Apps, Pid}) ->
+    {to_string(Subscription), ?_test(begin
+        ?assertEqual(ok, config:set("section_foo", "key_bar", "any", true)),
+        ?assertEqual({config_change,"section_foo", "key_bar", "any", true}, getmsg(Pid)),
+        ?assertEqual(ok, config:set("section_foo", "key_bar", "not_any", false)),
+        ?assertEqual({config_change,"section_foo", "key_bar", "not_any", false}, getmsg(Pid)),
+        ?assertEqual(ok, config:reload()),
+        ?assertEqual({config_change,"section_foo", "key_bar", "any", true}, getmsg(Pid)),
+        ok
+    end)}.
+
+should_notify_on_config_reload_flush(Subscription, {_Apps, Pid}) ->
+    {to_string(Subscription), ?_test(begin
+        ?assertEqual(ok, config:set("section_foo_temp", "key_bar", "any", false)),
+        ?assertEqual({config_change,"section_foo_temp", "key_bar", "any", false}, getmsg(Pid)),
+        ?assertEqual(ok, config:reload()),
+        ?assertEqual({config_change,"section_foo_temp", "key_bar", deleted, true}, getmsg(Pid)),
+        ok
+    end)}.
+
 
 spawn_config_listener() ->
     Self = self(),
@@ -758,3 +812,44 @@ n_notifiers() ->
 
 to_string(Term) ->
     lists:flatten(io_lib:format("~p", [Term])).
+
+with_process_restart(Name) ->
+    ok = stop_sync(whereis(Name), ?TIMEOUT),
+    Now = now_us(),
+    wait_process_restart(
+        Name, ?RESTART_TIMEOUT_IN_MILLISEC * 1000, 50, Now, Now).
+
+wait_process_restart(_Name, Timeout, _Delay, Started, Prev)
+        when Prev - Started > Timeout ->
+    timeout;
+wait_process_restart(Name, Timeout, Delay, Started, _Prev) ->
+    case whereis(Name) of
+        undefined ->
+            ok = timer:sleep(Delay),
+            wait_process_restart(Name, Timeout, Delay, Started, now_us());
+        Pid ->
+            Pid
+    end.
+
+stop_sync(Pid, Timeout) when is_pid(Pid) ->
+    MRef = erlang:monitor(process, Pid),
+    try
+        begin
+            catch unlink(Pid),
+            exit(Pid, kill),
+            receive
+            {'DOWN', MRef, _, _, _} ->
+                ok
+            after Timeout ->
+                timeout
+            end
+        end
+    after
+        erlang:demonitor(MRef, [flush])
+    end;
+stop_sync(_, _) -> error(badarg).
+
+
+now_us() ->
+    {MegaSecs, Secs, MicroSecs} = os:timestamp(),
+    (MegaSecs * 1000000 + Secs) * 1000000 + MicroSecs.
