@@ -38,6 +38,7 @@
     get_compacted_seq/1,
     get_compactor_pid/1,
     get_db_info/1,
+    get_partition_info/2,
     get_del_doc_count/1,
     get_doc_count/1,
     get_epochs/1,
@@ -57,15 +58,13 @@
     is_system_db/1,
     is_clustered/1,
     is_system_db_name/1,
+    is_partitioned/1,
 
     set_revs_limit/2,
     set_purge_infos_limit/2,
     set_security/2,
     set_user_ctx/2,
 
-    ensure_full_commit/1,
-    ensure_full_commit/2,
-
     load_validation_funs/1,
     reload_validation_funs/1,
 
@@ -85,6 +84,9 @@
     get_minimum_purge_seq/1,
     purge_client_exists/3,
 
+    validate_docid/2,
+    doc_from_json_obj_validate/2,
+
     update_doc/3,
     update_doc/4,
     update_docs/4,
@@ -185,11 +187,22 @@ reopen(#db{} = Db) ->
 incref(#db{} = Db) ->
     couch_db_engine:incref(Db).
 
-clustered_db(DbName, UserCtx) ->
-    clustered_db(DbName, UserCtx, []).
+clustered_db(DbName, Options) when is_list(Options) ->
+    UserCtx = couch_util:get_value(user_ctx, Options, #user_ctx{}),
+    SecProps = couch_util:get_value(security, Options, []),
+    Props = couch_util:get_value(props, Options, []),
+    {ok, #db{
+        name = DbName,
+        user_ctx = UserCtx,
+        security = SecProps,
+        options = [{props, Props}]
+    }};
+
+clustered_db(DbName, #user_ctx{} = UserCtx) ->
+    clustered_db(DbName, [{user_ctx, UserCtx}]).
 
 clustered_db(DbName, UserCtx, SecProps) ->
-    {ok, #db{name = DbName, user_ctx = UserCtx, security = SecProps}}.
+    clustered_db(DbName, [{user_ctx, UserCtx}, {security, SecProps}]).
 
 is_db(#db{}) ->
     true;
@@ -206,21 +219,16 @@ is_clustered(#db{}) ->
 is_clustered(?OLD_DB_REC = Db) ->
     ?OLD_DB_MAIN_PID(Db) == undefined.
 
-ensure_full_commit(#db{main_pid=Pid, instance_start_time=StartTime}) ->
-    ok = gen_server:call(Pid, full_commit, infinity),
-    {ok, StartTime}.
-
-ensure_full_commit(Db, RequiredSeq) ->
-    #db{main_pid=Pid, instance_start_time=StartTime} = Db,
-    ok = gen_server:call(Pid, {full_commit, RequiredSeq}, infinity),
-    {ok, StartTime}.
+is_partitioned(#db{options = Options}) ->
+    Props = couch_util:get_value(props, Options, []),
+    couch_util:get_value(partitioned, Props, false).
 
 close(#db{} = Db) ->
     ok = couch_db_engine:decref(Db);
 close(?OLD_DB_REC) ->
     ok.
 
-is_idle(#db{compactor_pid=nil, waiting_delayed_commit=nil} = Db) ->
+is_idle(#db{compactor_pid=nil} = Db) ->
     monitored_by(Db) == [];
 is_idle(_Db) ->
     false.
@@ -387,16 +395,18 @@ purge_docs(Db, IdRevs) ->
 -spec purge_docs(#db{}, [{UUId, Id, [Rev]}], [PurgeOption]) ->
     {ok, [Reply]} when
     UUId :: binary(),
-    Id :: binary(),
+    Id :: binary() | list(),
     Rev :: {non_neg_integer(), binary()},
     PurgeOption :: interactive_edit | replicated_changes,
     Reply :: {ok, []} | {ok, [Rev]}.
 purge_docs(#db{main_pid = Pid} = Db, UUIDsIdsRevs, Options) ->
+    UUIDsIdsRevs2 = [{UUID, couch_util:to_binary(Id), Revs}
+        || {UUID, Id, Revs}  <- UUIDsIdsRevs],
     % Check here if any UUIDs already exist when
     % we're not replicating purge infos
     IsRepl = lists:member(replicated_changes, Options),
     if IsRepl -> ok; true ->
-        UUIDs = [UUID || {UUID, _, _} <- UUIDsIdsRevs],
+        UUIDs = [UUID || {UUID, _, _} <- UUIDsIdsRevs2],
         lists:foreach(fun(Resp) ->
             if Resp == not_found -> ok; true ->
                 Fmt = "Duplicate purge info UIUD: ~s",
@@ -406,7 +416,7 @@ purge_docs(#db{main_pid = Pid} = Db, UUIDsIdsRevs, Options) ->
         end, get_purge_infos(Db, UUIDs))
     end,
     increment_stat(Db, [couchdb, database_purges]),
-    gen_server:call(Pid, {purge_docs, UUIDsIdsRevs, Options}).
+    gen_server:call(Pid, {purge_docs, UUIDsIdsRevs2, Options}).
 
 -spec get_purge_infos(#db{}, [UUId]) -> [PurgeInfo] when
     UUId :: binary(),
@@ -572,9 +582,6 @@ get_db_info(Db) ->
     {ok, DocCount} = get_doc_count(Db),
     {ok, DelDocCount} = get_del_doc_count(Db),
     SizeInfo = couch_db_engine:get_size_info(Db),
-    FileSize = couch_util:get_value(file, SizeInfo, null),
-    ActiveSize = couch_util:get_value(active, SizeInfo, null),
-    ExternalSize = couch_util:get_value(external, SizeInfo, null),
     DiskVersion = couch_db_engine:get_disk_version(Db),
     Uuid = case get_uuid(Db) of
         undefined -> null;
@@ -584,6 +591,10 @@ get_db_info(Db) ->
         undefined -> null;
         Else1 -> Else1
     end,
+    Props = case couch_db_engine:get_props(Db) of
+        undefined -> null;
+        Else2 -> {Else2}
+    end,
     InfoList = [
         {db_name, Name},
         {engine, couch_db_engine:get_engine(Db)},
@@ -593,22 +604,22 @@ get_db_info(Db) ->
         {purge_seq, couch_db_engine:get_purge_seq(Db)},
         {compact_running, Compactor /= nil},
         {sizes, {SizeInfo}},
-        % TODO: Remove this in 3.0
-        % These are legacy and have been duplicated under
-        % the sizes key since 2.0. We should make a note
-        % in our release notes that we'll remove these
-        % old versions in 3.0
-        {disk_size, FileSize}, % legacy
-        {data_size, ActiveSize},
-        {other, {[{data_size, ExternalSize}]}},
         {instance_start_time, StartTime},
         {disk_format_version, DiskVersion},
         {committed_update_seq, CommittedUpdateSeq},
         {compacted_seq, CompactedSeq},
+        {props, Props},
         {uuid, Uuid}
     ],
     {ok, InfoList}.
 
+get_partition_info(#db{} = Db, Partition) when is_binary(Partition) ->
+    Info = couch_db_engine:get_partition_info(Db, Partition),
+    {ok, Info};
+get_partition_info(_Db, _Partition) ->
+    throw({bad_request, <<"`partition` is not valid">>}).
+
+
 get_design_doc(#db{name = <<"shards/", _/binary>> = ShardDbName}, DDocId0) ->
     DDocId = couch_util:normalize_ddoc_id(DDocId0),
     DbName = mem3:dbname(ShardDbName),
@@ -730,9 +741,7 @@ get_security(?OLD_DB_REC = Db) ->
 set_security(#db{main_pid=Pid}=Db, {NewSecProps}) when is_list(NewSecProps) ->
     check_is_admin(Db),
     ok = validate_security_object(NewSecProps),
-    ok = gen_server:call(Pid, {set_security, NewSecProps}, infinity),
-    {ok, _} = ensure_full_commit(Db),
-    ok;
+    gen_server:call(Pid, {set_security, NewSecProps}, infinity);
 set_security(_, _) ->
     throw(bad_request).
 
@@ -782,6 +791,30 @@ name(#db{name=Name}) ->
 name(?OLD_DB_REC = Db) ->
     ?OLD_DB_NAME(Db).
 
+
+validate_docid(#db{} = Db, DocId) when is_binary(DocId) ->
+    couch_doc:validate_docid(DocId, name(Db)),
+    case is_partitioned(Db) of
+        true ->
+            couch_partition:validate_docid(DocId);
+        false ->
+            ok
+    end.
+
+
+doc_from_json_obj_validate(#db{} = Db, DocJson) ->
+    Doc = couch_doc:from_json_obj_validate(DocJson, name(Db)),
+    {Props} = DocJson,
+    case couch_util:get_value(<<"_id">>, Props) of
+        DocId when is_binary(DocId) ->
+            % Only validate the docid if it was provided
+            validate_docid(Db, DocId);
+        _ ->
+            ok
+    end,
+    Doc.
+
+
 update_doc(Db, Doc, Options) ->
     update_doc(Db, Doc, Options, interactive_edit).
 
@@ -831,7 +864,7 @@ group_alike_docs([Doc|Rest], [Bucket|RestBuckets]) ->
 
 validate_doc_update(#db{}=Db, #doc{id= <<"_design/",_/binary>>}=Doc, _GetDiskDocFun) ->
     case catch check_is_admin(Db) of
-        ok -> validate_ddoc(Db#db.name, Doc);
+        ok -> validate_ddoc(Db, Doc);
         Error -> Error
     end;
 validate_doc_update(#db{validate_doc_funs = undefined} = Db, Doc, Fun) ->
@@ -849,9 +882,9 @@ validate_doc_update(Db, Doc, GetDiskDocFun) ->
             validate_doc_update_int(Db, Doc, GetDiskDocFun)
     end.
 
-validate_ddoc(DbName, DDoc) ->
+validate_ddoc(Db, DDoc) ->
     try
-        ok = couch_index_server:validate(DbName, couch_doc:with_ejson_body(DDoc))
+        ok = couch_index_server:validate(Db, couch_doc:with_ejson_body(DDoc))
     catch
         throw:{invalid_design_doc, Reason} ->
             {bad_request, invalid_design_doc, Reason};
@@ -1148,7 +1181,7 @@ update_docs(Db, Docs0, Options, replicated_changes) ->
     end,
 
     {ok, DocBuckets, NonRepDocs, DocErrors}
-        = before_docs_update(Db, Docs, PrepValidateFun),
+        = before_docs_update(Db, Docs, PrepValidateFun, replicated_changes),
 
     DocBuckets2 = [[doc_flush_atts(Db, check_dup_atts(Doc))
             || Doc <- Bucket] || Bucket <- DocBuckets],
@@ -1166,7 +1199,7 @@ update_docs(Db, Docs0, Options, interactive_edit) ->
     end,
 
     {ok, DocBuckets, NonRepDocs, DocErrors}
-        = before_docs_update(Db, Docs, PrepValidateFun),
+        = before_docs_update(Db, Docs, PrepValidateFun, interactive_edit),
 
     if (AllOrNothing) and (DocErrors /= []) ->
         RefErrorDict = dict:from_list([{doc_tag(Doc), Doc} || Doc <- Docs]),
@@ -1209,24 +1242,6 @@ make_first_doc_on_disk(Db, Id, Pos, [{_Rev, #leaf{deleted=IsDel, ptr=Sp}} |_]=Do
     Revs = [Rev || {Rev, _} <- DocPath],
     make_doc(Db, Id, IsDel, Sp, {Pos, Revs}).
 
-set_commit_option(Options) ->
-    CommitSettings = {
-        [true || O <- Options, O==full_commit orelse O==delay_commit],
-        config:get("couchdb", "delayed_commits", "false")
-    },
-    case CommitSettings of
-    {[true], _} ->
-        Options; % user requested explicit commit setting, do not change it
-    {_, "true"} ->
-        Options; % delayed commits are enabled, do nothing
-    {_, "false"} ->
-        [full_commit|Options];
-    {_, Else} ->
-        couch_log:error("[couchdb] delayed_commits setting must be true/false,"
-                        " not ~p", [Else]),
-        [full_commit|Options]
-    end.
-
 collect_results_with_metrics(Pid, MRef, []) ->
     Begin = os:timestamp(),
     try
@@ -1252,14 +1267,12 @@ collect_results(Pid, MRef, ResultsAcc) ->
     end.
 
 write_and_commit(#db{main_pid=Pid, user_ctx=Ctx}=Db, DocBuckets1,
-        NonRepDocs, Options0) ->
+        NonRepDocs, Options) ->
     DocBuckets = prepare_doc_summaries(Db, DocBuckets1),
-    Options = set_commit_option(Options0),
     MergeConflicts = lists:member(merge_conflicts, Options),
-    FullCommit = lists:member(full_commit, Options),
     MRef = erlang:monitor(process, Pid),
     try
-        Pid ! {update_docs, self(), DocBuckets, NonRepDocs, MergeConflicts, FullCommit},
+        Pid ! {update_docs, self(), DocBuckets, NonRepDocs, MergeConflicts},
         case collect_results_with_metrics(Pid, MRef, []) of
         {ok, Results} -> {ok, Results};
         retry ->
@@ -1273,7 +1286,7 @@ write_and_commit(#db{main_pid=Pid, user_ctx=Ctx}=Db, DocBuckets1,
             % We only retry once
             DocBuckets3 = prepare_doc_summaries(Db2, DocBuckets2),
             close(Db2),
-            Pid ! {update_docs, self(), DocBuckets3, NonRepDocs, MergeConflicts, FullCommit},
+            Pid ! {update_docs, self(), DocBuckets3, NonRepDocs, MergeConflicts},
             case collect_results_with_metrics(Pid, MRef, []) of
             {ok, Results} -> {ok, Results};
             retry -> throw({update_error, compaction_retry})
@@ -1301,7 +1314,7 @@ prepare_doc_summaries(Db, BucketList) ->
                 meta = [
                     {size_info, SizeInfo},
                     {atts_stream, AttsStream},
-                    {ejson_size, ?term_size(Body)}
+                    {ejson_size, couch_ejson_size:encoded_size(Body)}
                 ] ++ Doc0#doc.meta
             },
             couch_db_engine:serialize_doc(Db, Doc1)
@@ -1309,7 +1322,7 @@ prepare_doc_summaries(Db, BucketList) ->
         Bucket) || Bucket <- BucketList].
 
 
-before_docs_update(#db{validate_doc_funs = VDFuns} = Db, Docs, PVFun) ->
+before_docs_update(#db{validate_doc_funs = VDFuns} = Db, Docs, PVFun, UpdateType) ->
     increment_stat(Db, [couchdb, database_writes]),
 
     % Separate _local docs from normal docs
@@ -1324,7 +1337,7 @@ before_docs_update(#db{validate_doc_funs = VDFuns} = Db, Docs, PVFun) ->
     DocBuckets = lists:map(fun(Bucket) ->
         lists:map(fun(Doc) ->
             DocWithBody = couch_doc:with_ejson_body(Doc),
-            couch_db_plugin:before_doc_update(Db, DocWithBody)
+            couch_db_plugin:before_doc_update(Db, DocWithBody, UpdateType)
         end, Bucket)
     end, BucketList),
 
@@ -1468,6 +1481,17 @@ calculate_start_seq(_Db, _Node, Seq) when is_integer(Seq) ->
 calculate_start_seq(Db, Node, {Seq, Uuid}) ->
     % Treat the current node as the epoch node
     calculate_start_seq(Db, Node, {Seq, Uuid, Node});
+calculate_start_seq(Db, _Node, {Seq, {split, Uuid}, EpochNode}) ->
+    case is_owner(EpochNode, Seq, get_epochs(Db)) of
+        true ->
+            % Find last replicated sequence from split source to target
+            mem3_rep:find_split_target_seq(Db, EpochNode, Uuid, Seq);
+        false ->
+            couch_log:warning("~p calculate_start_seq not owner "
+                "db: ~p, seq: ~p, uuid: ~p, epoch_node: ~p, epochs: ~p",
+                [?MODULE, Db#db.name, Seq, Uuid, EpochNode, get_epochs(Db)]),
+            0
+    end;
 calculate_start_seq(Db, _Node, {Seq, Uuid, EpochNode}) ->
     case is_prefix(Uuid, get_uuid(Db)) of
         true ->
@@ -1491,7 +1515,14 @@ calculate_start_seq(Db, _Node, {Seq, Uuid, EpochNode}) ->
 calculate_start_seq(Db, _Node, {replace, OriginalNode, Uuid, Seq}) ->
     case is_prefix(Uuid, couch_db:get_uuid(Db)) of
         true ->
-            start_seq(get_epochs(Db), OriginalNode, Seq);
+            try
+                start_seq(get_epochs(Db), OriginalNode, Seq)
+            catch throw:epoch_mismatch ->
+                couch_log:warning("~p start_seq duplicate uuid on node: ~p "
+                    "db: ~p, seq: ~p, uuid: ~p, epoch_node: ~p",
+                    [?MODULE, node(), Db#db.name, Seq, Uuid, OriginalNode]),
+                0
+            end;
         false ->
             {replace, OriginalNode, Uuid, Seq}
     end.
@@ -1538,8 +1569,8 @@ start_seq([{_, NewSeq}, {OrigNode, _} | _], OrigNode, Seq) when Seq > NewSeq ->
     NewSeq;
 start_seq([_ | Rest], OrigNode, Seq) ->
     start_seq(Rest, OrigNode, Seq);
-start_seq([], OrigNode, Seq) ->
-    erlang:error({epoch_mismatch, OrigNode, Seq}).
+start_seq([], _OrigNode, _Seq) ->
+    throw(epoch_mismatch).
 
 
 fold_docs(Db, UserFun, UserAcc) ->
@@ -1767,7 +1798,7 @@ is_system_db_name(DbName) when is_binary(DbName) ->
     Suffix = filename:basename(Normalized),
     case {filename:dirname(Normalized), lists:member(Suffix, ?SYSTEM_DATABASES)} of
         {<<".">>, Result} -> Result;
-        {Prefix, false} -> false;
+        {_Prefix, false} -> false;
         {Prefix, true} ->
             ReOpts =  [{capture,none}, dollar_endonly],
             re:run(Prefix, ?DBNAME_REGEX, ReOpts) == match
@@ -1839,13 +1870,19 @@ set_design_doc_end_key(Options, rev) ->
 -ifdef(TEST).
 -include_lib("eunit/include/eunit.hrl").
 
-setup() ->
+setup_all() ->
     ok = meck:new(couch_epi, [passthrough]),
     ok = meck:expect(couch_epi, decide, fun(_, _, _, _, _) -> no_decision end),
     ok.
 
+teardown_all(_) ->
+    meck:unload().
+
+setup() ->
+    meck:reset([couch_epi]).
+
 teardown(_) ->
-    (catch meck:unload(couch_epi)).
+    ok.
 
 validate_dbname_success_test_() ->
     Cases =
@@ -1855,8 +1892,15 @@ validate_dbname_success_test_() ->
             [generate_cases_with_shards(?b2l(SystemDb))
                 || SystemDb <- ?SYSTEM_DATABASES]),
     {
-        foreach, fun setup/0, fun teardown/1,
-        [should_pass_validate_dbname(A) || {_, A} <- Cases]
+        setup,
+        fun setup_all/0,
+        fun teardown_all/1,
+        {
+            foreach,
+            fun setup/0,
+            fun teardown/1,
+            [should_pass_validate_dbname(A) || {_, A} <- Cases]
+        }
     }.
 
 validate_dbname_fail_test_() ->
@@ -1867,8 +1911,15 @@ validate_dbname_fail_test_() ->
        ++ generate_cases("!abcdefg/werwej/_users")
        ++ generate_cases_with_shards("!abcdefg/werwej/_users"),
     {
-        foreach, fun setup/0, fun teardown/1,
-        [should_fail_validate_dbname(A) || {_, A} <- Cases]
+        setup,
+        fun setup_all/0,
+        fun teardown_all/1,
+        {
+            foreach,
+            fun setup/0,
+            fun teardown/1,
+            [should_fail_validate_dbname(A) || {_, A} <- Cases]
+        }
     }.
 
 normalize_dbname_test_() ->
@@ -1910,18 +1961,24 @@ should_fail_validate_dbname(DbName) ->
 
 calculate_start_seq_test_() ->
     {
-        foreach,
-        fun setup_start_seq/0,
-        fun teardown_start_seq/1,
-        [
-            t_calculate_start_seq_uuid_mismatch(),
-            t_calculate_start_seq_is_owner(),
-            t_calculate_start_seq_not_owner(),
-            t_calculate_start_seq_raw()
-        ]
+        setup,
+        fun setup_start_seq_all/0,
+        fun teardown_start_seq_all/1,
+        {
+            foreach,
+            fun setup_start_seq/0,
+            fun teardown_start_seq/1,
+            [
+                t_calculate_start_seq_uuid_mismatch(),
+                t_calculate_start_seq_is_owner(),
+                t_calculate_start_seq_not_owner(),
+                t_calculate_start_seq_raw(),
+                t_calculate_start_seq_epoch_mismatch()
+            ]
+        }
     }.
 
-setup_start_seq() ->
+setup_start_seq_all() ->
     meck:new(couch_db_engine, [passthrough]),
     meck:expect(couch_db_engine, get_uuid, fun(_) -> <<"foo">> end),
     ok = meck:expect(couch_log, warning, 2, ok),
@@ -1931,9 +1988,18 @@ setup_start_seq() ->
     ],
     meck:expect(couch_db_engine, get_epochs, fun(_) -> Epochs end).
 
-teardown_start_seq(_) ->
+teardown_start_seq_all(_) ->
     meck:unload().
 
+setup_start_seq() ->
+    meck:reset([
+        couch_db_engine,
+        couch_log
+    ]).
+
+teardown_start_seq(_) ->
+    ok.
+
 t_calculate_start_seq_uuid_mismatch() ->
     ?_test(begin
         Db = test_util:fake_db([]),
@@ -1962,6 +2028,14 @@ t_calculate_start_seq_raw() ->
         ?assertEqual(13, Seq)
     end).
 
+t_calculate_start_seq_epoch_mismatch() ->
+    ?_test(begin
+        Db = test_util:fake_db([]),
+        SeqIn = {replace, not_this_node, get_uuid(Db), 42},
+        Seq = calculate_start_seq(Db, node1, SeqIn),
+        ?assertEqual(0, Seq)
+    end).
+
 is_owner_test() ->
     ?assertNot(is_owner(foo, 1, [])),
     ?assertNot(is_owner(foo, 1, [{foo, 1}])),
