@@ -21,7 +21,8 @@
     get_stale_proc_count/0,
     new_proc/1,
     reload/0,
-    terminate_stale_procs/0
+    terminate_stale_procs/0,
+    get_servers_from_env/1
 ]).
 
 -export([
