@@ -377,17 +377,7 @@ code_change(_OldVsn, State, _Extra) ->
 
 maybe_restart_updater(#st{waiters=[]}) ->
     ok;
-maybe_restart_updater(#st{mod=Mod, idx_state=IdxState}=State) ->
-    couch_util:with_db(Mod:get(db_name, IdxState), fun(Db) ->
-        UpdateSeq = couch_db:get_update_seq(Db),
-        CommittedSeq = couch_db:get_committed_update_seq(Db),
-        CanUpdate = UpdateSeq > CommittedSeq,
-        UOpts = Mod:get(update_options, IdxState),
-        case CanUpdate and lists:member(committed_only, UOpts) of
-            true -> couch_db:ensure_full_commit(Db);
-            false -> ok
-        end
-    end),
+maybe_restart_updater(#st{idx_state=IdxState}=State) ->
     couch_index_updater:run(State#st.updater, IdxState).
 
 
@@ -480,18 +470,25 @@ get(idx_name, _, _) ->
 get(signature, _, _) ->
     <<61,237,157,230,136,93,96,201,204,17,137,186,50,249,44,135>>.
 
-setup(Settings) ->
+setup_all() ->
+    Ctx = test_util:start_couch(),
     ok = meck:new([config], [passthrough]),
     ok = meck:new([test_index], [non_strict]),
+    ok = meck:expect(test_index, get, fun get/3),
+    Ctx.
+
+teardown_all(Ctx) ->
+    meck:unload(),
+    test_util:stop_couch(Ctx).
+
+setup(Settings) ->
+    meck:reset([config, test_index]),
     ok = meck:expect(config, get, fun(Section, Key) ->
         configure(Section, Key, Settings)
     end),
-    ok = meck:expect(test_index, get, fun get/3),
     {undefined, #st{mod = {test_index}}}.
 
 teardown(_, _) ->
-    (catch meck:unload(config)),
-    (catch meck:unload(test_index)),
     ok.
 
 configure("view_compaction", "enabled_recompaction", [Global, _Db, _Index]) ->
@@ -508,10 +505,12 @@ recompaction_configuration_test_() ->
         "Compaction tests",
         {
             setup,
-            fun test_util:start_couch/0, fun test_util:stop_couch/1,
+            fun setup_all/0,
+            fun teardown_all/1,
             {
                 foreachx,
-                fun setup/1, fun teardown/2,
+                fun setup/1,
+                fun teardown/2,
                 recompaction_configuration_tests()
             }
         }
