@@ -18,9 +18,8 @@
     parse_rep_db/3,
     parse_rep_doc_without_id/1,
     parse_rep_doc_without_id/2,
-    before_doc_update/2,
+    before_doc_update/3,
     after_doc_read/2,
-    ensure_rep_db_exists/0,
     ensure_rep_ddoc_exists/1,
     ensure_cluster_rep_ddoc_exists/1,
     remove_state_fields/2,
@@ -121,20 +120,6 @@ update_error(#rep{db_name = DbName, doc_id = DocId, id = RepId}, Error) ->
     ok.
 
 
--spec ensure_rep_db_exists() -> {ok, Db::any()}.
-ensure_rep_db_exists() ->
-    Db = case couch_db:open_int(?REP_DB_NAME, [?CTX, sys_db,
-            nologifmissing]) of
-        {ok, Db0} ->
-            Db0;
-        _Error ->
-            {ok, Db0} = couch_db:create(?REP_DB_NAME, [?CTX, sys_db]),
-            Db0
-    end,
-    ok = ensure_rep_ddoc_exists(?REP_DB_NAME),
-    {ok, Db}.
-
-
 -spec ensure_rep_ddoc_exists(binary()) -> ok.
 ensure_rep_ddoc_exists(RepDb) ->
     case mem3:belongs(RepDb, ?REP_DESIGN_DOC) of
@@ -256,15 +241,15 @@ parse_rep_doc(Doc, UserCtx) ->
 
 -spec parse_rep_doc_without_id({[_]}, #user_ctx{}) -> {ok, #rep{}}.
 parse_rep_doc_without_id({Props}, UserCtx) ->
-    Proxy = get_value(<<"proxy">>, Props, <<>>),
+    {SrcProxy, TgtProxy} = parse_proxy_settings(Props),
     Opts = make_options(Props),
     case get_value(cancel, Opts, false) andalso
         (get_value(id, Opts, nil) =/= nil) of
     true ->
         {ok, #rep{options = Opts, user_ctx = UserCtx}};
     false ->
-        Source = parse_rep_db(get_value(<<"source">>, Props), Proxy, Opts),
-        Target = parse_rep_db(get_value(<<"target">>, Props), Proxy, Opts),
+        Source = parse_rep_db(get_value(<<"source">>, Props), SrcProxy, Opts),
+        Target = parse_rep_db(get_value(<<"target">>, Props), TgtProxy, Opts),
         {Type, View} = case couch_replicator_filters:view_type(Props, Opts) of
         {error, Error} ->
             throw({bad_request, Error});
@@ -291,6 +276,25 @@ parse_rep_doc_without_id({Props}, UserCtx) ->
     end.
 
 
+parse_proxy_settings(Props) when is_list(Props) ->
+    Proxy = get_value(<<"proxy">>, Props, <<>>),
+    SrcProxy = get_value(<<"source_proxy">>, Props, <<>>),
+    TgtProxy = get_value(<<"target_proxy">>, Props, <<>>),
+
+    case Proxy =/= <<>> of
+        true when SrcProxy =/= <<>> ->
+            Error = "`proxy` is mutually exclusive with `source_proxy`",
+            throw({bad_request, Error});
+        true when  TgtProxy =/= <<>> ->
+            Error = "`proxy` is mutually exclusive with `target_proxy`",
+            throw({bad_request, Error});
+        true ->
+            {Proxy, Proxy};
+        false ->
+            {SrcProxy, TgtProxy}
+    end.
+
+
 % Update a #rep{} record with a replication_id. Calculating the id might involve
 % fetching a filter from the source db, and so it could fail intermetently.
 % In case of a failure to fetch the filter this function will throw a
@@ -423,8 +427,8 @@ parse_rep_db(<<"http://", _/binary>> = Url, Proxy, Options) ->
 parse_rep_db(<<"https://", _/binary>> = Url, Proxy, Options) ->
     parse_rep_db({[{<<"url">>, Url}]}, Proxy, Options);
 
-parse_rep_db(<<DbName/binary>>, _Proxy, _Options) ->
-    DbName;
+parse_rep_db(<<_/binary>>, _Proxy, _Options) ->
+    throw({error, local_endpoints_not_supported});
 
 parse_rep_db(undefined, _Proxy, _Options) ->
     throw({error, <<"Missing replicator database">>}).
@@ -615,10 +619,10 @@ ssl_verify_options(false) ->
     [{verify, verify_none}].
 
 
--spec before_doc_update(#doc{}, Db::any()) -> #doc{}.
-before_doc_update(#doc{id = <<?DESIGN_DOC_PREFIX, _/binary>>} = Doc, _Db) ->
+-spec before_doc_update(#doc{}, Db::any(), couch_db:update_type()) -> #doc{}.
+before_doc_update(#doc{id = <<?DESIGN_DOC_PREFIX, _/binary>>} = Doc, _Db, _UpdateType) ->
     Doc;
-before_doc_update(#doc{body = {Body}} = Doc, Db) ->
+before_doc_update(#doc{body = {Body}} = Doc, Db, _UpdateType) ->
     #user_ctx{
        roles = Roles,
        name = Name
@@ -683,8 +687,12 @@ strip_credentials(Url) when is_binary(Url) ->
         "http(s)?://(?:[^:]+):[^@]+@(.*)$",
         "http\\1://\\2",
         [{return, binary}]);
-strip_credentials({Props}) ->
-    {lists:keydelete(<<"headers">>, 1, Props)}.
+strip_credentials({Props0}) ->
+    Props1 = lists:keydelete(<<"headers">>, 1, Props0),
+    % Strip "auth" just like headers, for replication plugins it can be a place
+    % to stash credential that are not necessarily in headers
+    Props2 = lists:keydelete(<<"auth">>, 1, Props1),
+    {Props2}.
 
 
 error_reason({shutdown, Error}) ->
@@ -773,6 +781,10 @@ check_strip_credentials_test() ->
         {
             {[{<<"_id">>, <<"foo">>}]},
             {[{<<"_id">>, <<"foo">>}, {<<"headers">>, <<"baz">>}]}
+        },
+        {
+            {[{<<"_id">>, <<"foo">>}]},
+            {[{<<"_id">>, <<"foo">>}, {<<"auth">>, <<"pluginsecret">>}]}
         }
     ]].
 
@@ -797,8 +809,7 @@ create_vdu(DbName) ->
             id = <<"_design/vdu">>,
             body = {[{<<"validate_doc_update">>, VduFun}]}
         },
-        {ok, _} = couch_db:update_docs(Db, [Doc]),
-        couch_db:ensure_full_commit(Db)
+        {ok, _} = couch_db:update_docs(Db, [Doc])
     end).
 
 
@@ -822,4 +833,29 @@ t_vdu_does_not_crash_on_save(DbName) ->
         ?assertEqual({ok, forbidden}, save_rep_doc(DbName, Doc))
     end).
 
+
+local_replication_endpoint_error_test_() ->
+     {
+        foreach,
+        fun () -> meck:expect(config, get,
+            fun(_, _, Default) -> Default end)
+        end,
+        fun (_) -> meck:unload() end,
+        [
+            t_error_on_local_endpoint()
+        ]
+    }.
+
+
+t_error_on_local_endpoint() ->
+    ?_test(begin
+        RepDoc = {[
+            {<<"_id">>, <<"someid">>},
+            {<<"source">>, <<"localdb">>},
+            {<<"target">>, <<"http://somehost.local/tgt">>}
+        ]},
+        Expect = local_endpoints_not_supported,
+        ?assertThrow({bad_rep_doc, Expect}, parse_rep_doc_without_id(RepDoc))
+    end).
+
 -endif.
