@@ -79,7 +79,7 @@ run(Dir) ->
 app(Name, Modules) ->
     App = {application, Name,
            [{description, atom_to_list(Name)},
-            {vsn, "0c435f3"},
+            {vsn, "3.1.2"},
             {modules, Modules},
             {registered, []},
             {applications, [kernel, stdlib]}]},
