@@ -75,7 +75,7 @@ is_single_node_enabled(Dbs) ->
     end.
 
 cluster_system_dbs() ->
-    ["_users", "_replicator", "_global_changes"].
+    ["_users", "_replicator"].
 
 
 has_cluster_system_dbs([]) ->
@@ -200,6 +200,8 @@ setup_node(NewCredentials, NewBindAddress, NodeCount, Port) ->
 finish_cluster(Options) ->
     ok = wait_connected(),
     ok = sync_admins(),
+    ok = sync_uuid(),
+    ok = sync_auth_secret(),
     Dbs = proplists:get_value(ensure_dbs_exist, Options, cluster_system_dbs()),
     finish_cluster_int(Dbs, has_cluster_system_dbs(Dbs)).
 
@@ -241,13 +243,26 @@ sync_admins() ->
 
 
 sync_admin(User, Pass) ->
+    sync_config("admins", User, Pass).
+
+
+sync_uuid() ->
+    Uuid = config:get("couchdb", "uuid"),
+    sync_config("couchdb", "uuid", Uuid).
+
+sync_auth_secret() ->
+    Secret = config:get("couch_httpd_auth", "secret"),
+    sync_config("couch_httpd_auth", "secret", Secret).
+
+
+sync_config(Section, Key, Value) ->
     {Results, Errors} = rpc:multicall(other_nodes(), config, set,
-        ["admins", User, Pass]),
+        [Section, Key, Value]),
     case validate_multicall(Results, Errors) of
         ok ->
             ok;
         error ->
-            log:error("~p sync_admin results ~p errors ~p",
+            couch_log:error("~p sync_admin results ~p errors ~p",
                 [?MODULE, Results, Errors]),
             Reason = "Cluster setup unable to sync admin passwords",
             throw({setup_error, Reason})
