@@ -11,6 +11,19 @@
 // the License.
 //
 
+/*
+ * Quit current test execution if it is tagged as skipped or ported to elixir
+ */
+function quitIfSkippedOrPorted() {
+  if(couchTests.skip) {
+    quit(2);
+  }
+
+  if(couchTests.elixir) {
+    quit(3);
+  }
+}
+
 /*
  * Futon test suite was designed to be able to run all tests populated into
  * couchTests. Here we should only be loading one test, so we'll pop the first
@@ -43,6 +56,8 @@ function runTest() {
   }
 }
 
+quitIfSkippedOrPorted();
+
 waitForSuccess(CouchDB.isRunning, 'isRunning');
 
 runTest();
