@@ -11,7 +11,10 @@
 // the License.
 
 // Do some basic tests.
+couchTests.elixir = true;
 couchTests.form_submit = function(debug) {
+    return console.log('done in test/elixir/test/form_summit_test.exs');
+
     var db_name = get_random_db_name();
     var db = new CouchDB(db_name, {"X-Couch-Full-Commit":"false"});
     db.createDb();
